@@ -1,54 +1,50 @@
-from __future__ import unicode_literals
-from __future__ import print_function
+from __future__ import print_function, unicode_literals
 
-import os
-import sys
-import traceback
+import itertools
 import logging
+import os
+import re
+import shutil
+import sys
 import threading
-from time import time
+import traceback
+from collections import namedtuple
 from datetime import datetime
 from io import open
-from collections import namedtuple
 from sqlite3 import OperationalError, sqlite_version
-import shutil
+from time import time
 
-from cli_helpers.tabular_output import TabularOutputFormatter
-from cli_helpers.tabular_output import preprocessors
 import click
 import sqlparse
+from cli_helpers.tabular_output import TabularOutputFormatter, preprocessors
+from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
 from prompt_toolkit.completion import DynamicCompleter
-from prompt_toolkit.enums import DEFAULT_BUFFER, EditingMode
-from prompt_toolkit.shortcuts import PromptSession, CompleteStyle
-from prompt_toolkit.styles.pygments import style_from_pygments_cls
 from prompt_toolkit.document import Document
+from prompt_toolkit.enums import DEFAULT_BUFFER, EditingMode
 from prompt_toolkit.filters import HasFocus, IsDone
 from prompt_toolkit.formatted_text import ANSI
+from prompt_toolkit.history import FileHistory
 from prompt_toolkit.layout.processors import (
-    HighlightMatchingBracketProcessor,
     ConditionalProcessor,
+    HighlightMatchingBracketProcessor,
 )
 from prompt_toolkit.lexers import PygmentsLexer
-from prompt_toolkit.history import FileHistory
-from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
+from prompt_toolkit.shortcuts import CompleteStyle, PromptSession
 
-from .packages.special.main import NO_QUERY
-from .packages.prompt_utils import confirm, confirm_destructive_query
-from .packages import special
-from .sqlcompleter import SQLCompleter
-from .clitoolbar import create_toolbar_tokens_func
-from .clistyle import style_factory, style_factory_output
-from .sqlexecute import SQLExecute
+from .__init__ import __version__
 from .clibuffer import cli_is_multiline
+from .clistyle import style_factory, style_factory_output
+from .clitoolbar import create_toolbar_tokens_func
 from .completion_refresher import CompletionRefresher
 from .config import config_location, ensure_dir_exists, get_config
 from .key_bindings import cli_bindings
-from .encodingutils import utf8tounicode, text_type
 from .lexer import LiteCliLexer
-from .__init__ import __version__
+from .packages import special
 from .packages.filepaths import dir_path_exists
-
-import itertools
+from .packages.prompt_utils import confirm, confirm_destructive_query
+from .packages.special.main import NO_QUERY
+from .sqlcompleter import SQLCompleter
+from .sqlexecute import SQLExecute
 
 click.disable_unicode_literals_warning = True
 
@@ -94,9 +90,7 @@ class LiteCli(object):
         self.login_path_as_host = c["main"].as_bool("login_path_as_host")
 
         # read from cli argument or user config file
-        self.auto_vertical_output = auto_vertical_output or c["main"].as_bool(
-            "auto_vertical_output"
-        )
+        self.auto_vertical_output = auto_vertical_output or c["main"].as_bool("auto_vertical_output")
 
         # audit log
         if self.logfile is None and "audit_log" in c["main"]:
@@ -112,9 +106,7 @@ class LiteCli(object):
         # Load startup commands.
         try:
             self.startup_commands = c["startup_commands"]
-        except (
-            KeyError
-        ):  # Redundant given the load_config() function that merges in the standard config, but put here to avoid fail if user do not have updated config file.
+        except KeyError:  # Redundant given the load_config() function that merges in the standard config, but put here to avoid fail if user do not have updated config file.
             self.startup_commands = None
 
         self.completion_refresher = CompletionRefresher()
@@ -123,9 +115,7 @@ class LiteCli(object):
         self.initialize_logging()
 
         prompt_cnf = self.read_my_cnf_files(["prompt"])["prompt"]
-        self.prompt_format = (
-            prompt or prompt_cnf or c["main"]["prompt"] or self.default_prompt
-        )
+        self.prompt_format = prompt or prompt_cnf or c["main"]["prompt"] or self.default_prompt
         self.prompt_continuation_format = c["main"]["prompt_continuation"]
         keyword_casing = c["main"].get("keyword_casing", "auto")
 
@@ -270,10 +260,7 @@ class LiteCli(object):
             )
             return
 
-        formatter = logging.Formatter(
-            "%(asctime)s (%(process)d/%(threadName)s) "
-            "%(name)s %(levelname)s - %(message)s"
-        )
+        formatter = logging.Formatter("%(asctime)s (%(process)d/%(threadName)s) " "%(name)s %(levelname)s - %(message)s")
 
         handler.setFormatter(formatter)
 
@@ -371,8 +358,7 @@ class LiteCli(object):
         else:
             history = None
             self.echo(
-                'Error: Unable to open the history file "{}". '
-                "Your query history will not be saved.".format(history_file),
+                'Error: Unable to open the history file "{}". ' "Your query history will not be saved.".format(history_file),
                 err=True,
                 fg="red",
             )
@@ -385,10 +371,7 @@ class LiteCli(object):
 
         def get_message():
             prompt = self.get_prompt(self.prompt_format)
-            if (
-                self.prompt_format == self.default_prompt
-                and len(prompt) > self.max_len_prompt
-            ):
+            if self.prompt_format == self.default_prompt and len(prompt) > self.max_len_prompt:
                 prompt = self.get_prompt("\\d> ")
             prompt = prompt.replace("\\x1b", "\x1b")
             return ANSI(prompt)
@@ -400,6 +383,47 @@ class LiteCli(object):
         def show_suggestion_tip():
             return iterations < 2
 
+        def output_res(res, start):
+            result_count = 0
+            mutating = False
+            for title, cur, headers, status in res:
+                logger.debug("headers: %r", headers)
+                logger.debug("rows: %r", cur)
+                logger.debug("status: %r", status)
+                threshold = 1000
+                if is_select(status) and cur and cur.rowcount > threshold:
+                    self.echo(
+                        "The result set has more than {} rows.".format(threshold),
+                        fg="red",
+                    )
+                    if not confirm("Do you want to continue?"):
+                        self.echo("Aborted!", err=True, fg="red")
+                        break
+
+                if self.auto_vertical_output:
+                    max_width = self.prompt_app.output.get_size().columns
+                else:
+                    max_width = None
+
+                formatted = self.format_output(title, cur, headers, special.is_expanded_output(), max_width)
+
+                t = time() - start
+                try:
+                    if result_count > 0:
+                        self.echo("")
+                    try:
+                        self.output(formatted, status)
+                    except KeyboardInterrupt:
+                        pass
+                    self.echo("Time: %0.03fs" % t)
+                except KeyboardInterrupt:
+                    pass
+
+                start = time()
+                result_count += 1
+                mutating = mutating or is_mutating(status)
+            return mutating
+
         def one_iteration(text=None):
             if text is None:
                 try:
@@ -417,6 +441,24 @@ class LiteCli(object):
                     self.echo(str(e), err=True, fg="red")
                     return
 
+                if special.is_llm_command(text):
+                    try:
+                        start = time()
+                        cur = self.sqlexecute.conn.cursor()
+                        context, sql = special.handle_llm(text, cur)
+                        if context:
+                            click.echo(context)
+                        text = self.prompt_app.prompt(default=sql)
+                    except KeyboardInterrupt:
+                        return
+                    except special.FinishIteration as e:
+                        return output_res(e.results, start) if e.results else None
+                    except RuntimeError as e:
+                        logger.error("sql: %r, error: %r", text, e)
+                        logger.error("traceback: %r", traceback.format_exc())
+                        self.echo(str(e), err=True, fg="red")
+                        return
+
             if not text.strip():
                 return
 
@@ -430,9 +472,6 @@ class LiteCli(object):
                     self.echo("Wise choice!")
                     return
 
-            # Keep track of whether or not the query is mutating. In case
-            # of a multi-statement query, the overall query is considered
-            # mutating if any one of the component statements is mutating
             mutating = False
 
             try:
@@ -449,46 +488,12 @@ class LiteCli(object):
                 res = sqlexecute.run(text)
                 self.formatter.query = text
                 successful = True
-                result_count = 0
-                for title, cur, headers, status in res:
-                    logger.debug("headers: %r", headers)
-                    logger.debug("rows: %r", cur)
-                    logger.debug("status: %r", status)
-                    threshold = 1000
-                    if is_select(status) and cur and cur.rowcount > threshold:
-                        self.echo(
-                            "The result set has more than {} rows.".format(threshold),
-                            fg="red",
-                        )
-                        if not confirm("Do you want to continue?"):
-                            self.echo("Aborted!", err=True, fg="red")
-                            break
-
-                    if self.auto_vertical_output:
-                        max_width = self.prompt_app.output.get_size().columns
-                    else:
-                        max_width = None
-
-                    formatted = self.format_output(
-                        title, cur, headers, special.is_expanded_output(), max_width
-                    )
-
-                    t = time() - start
-                    try:
-                        if result_count > 0:
-                            self.echo("")
-                        try:
-                            self.output(formatted, status)
-                        except KeyboardInterrupt:
-                            pass
-                        self.echo("Time: %0.03fs" % t)
-                    except KeyboardInterrupt:
-                        pass
-
-                    start = time()
-                    result_count += 1
-                    mutating = mutating or is_mutating(status)
                 special.unset_once_if_written()
+                # Keep track of whether or not the query is mutating. In case
+                # of a multi-statement query, the overall query is considered
+                # mutating if any one of the component statements is mutating
+                mutating = output_res(res, start)
+                special.unset_pipe_once_if_written()
             except EOFError as e:
                 raise e
             except KeyboardInterrupt:
@@ -601,9 +606,7 @@ class LiteCli(object):
                             for title, cur, headers, status in res:
                                 if title == "dot command not implemented":
                                     self.echo(
-                                        "The SQLite dot command '"
-                                        + command.split(" ", 1)[0]
-                                        + "' is not yet implemented.",
+                                        "The SQLite dot command '" + command.split(" ", 1)[0] + "' is not yet implemented.",
                                         fg="yellow",
                                     )
                                 else:
@@ -619,9 +622,7 @@ class LiteCli(object):
         try:
             startup_commands()
         except Exception as e:
-            self.echo(
-                "Could not execute all startup commands: \n" + str(e), fg="yellow"
-            )
+            self.echo("Could not execute all startup commands: \n" + str(e), fg="yellow")
 
         try:
             while True:
@@ -635,7 +636,7 @@ class LiteCli(object):
     def log_output(self, output):
         """Log the output in the audit log, if it's enabled."""
         if self.logfile:
-            click.echo(utf8tounicode(output), file=self.logfile)
+            click.echo(output, file=self.logfile)
 
     def echo(self, s, **kwargs):
         """Print a message to stdout.
@@ -651,11 +652,7 @@ class LiteCli(object):
     def get_output_margin(self, status=None):
         """Get the output margin (number of rows for the prompt, footer and
         timing message."""
-        margin = (
-            self.get_reserved_space()
-            + self.get_prompt(self.prompt_format).count("\n")
-            + 2
-        )
+        margin = self.get_reserved_space() + self.get_prompt(self.prompt_format).count("\n") + 2
         if status:
             margin += 1 + status.count("\n")
 
@@ -683,6 +680,7 @@ class LiteCli(object):
                 self.log_output(line)
                 special.write_tee(line)
                 special.write_once(line)
+                special.write_pipe_once(line)
 
                 if fits or output_via_pager:
                     # buffering
@@ -741,9 +739,7 @@ class LiteCli(object):
             },
         )
 
-        return [
-            (None, None, None, "Auto-completion refresh started in the background.")
-        ]
+        return [(None, None, None, "Auto-completion refresh started in the background.")]
 
     def _on_completions_refreshed(self, new_completer):
         """Swap the completer object in cli with the newly created completer."""
@@ -757,25 +753,35 @@ class LiteCli(object):
 
     def get_completions(self, text, cursor_positition):
         with self._completer_lock:
-            return self.completer.get_completions(
-                Document(text=text, cursor_position=cursor_positition), None
-            )
+            return self.completer.get_completions(Document(text=text, cursor_position=cursor_positition), None)
 
     def get_prompt(self, string):
-        self.logger.debug("Getting prompt")
+        self.logger.debug("Getting prompt %r", string)
         sqlexecute = self.sqlexecute
         now = datetime.now()
-        string = string.replace("\\d", sqlexecute.dbname or "(none)")
-        string = string.replace("\\f", os.path.basename(sqlexecute.dbname or "(none)"))
-        string = string.replace("\\n", "\n")
-        string = string.replace("\\D", now.strftime("%a %b %d %H:%M:%S %Y"))
-        string = string.replace("\\m", now.strftime("%M"))
-        string = string.replace("\\P", now.strftime("%p"))
-        string = string.replace("\\R", now.strftime("%H"))
-        string = string.replace("\\r", now.strftime("%I"))
-        string = string.replace("\\s", now.strftime("%S"))
-        string = string.replace("\\_", " ")
-        return string
+
+        # Prepare the replacements dictionary
+        replacements = {
+            r"\d": sqlexecute.dbname or "(none)",
+            r"\f": os.path.basename(sqlexecute.dbname or "(none)"),
+            r"\n": "\n",
+            r"\D": now.strftime("%a %b %d %H:%M:%S %Y"),
+            r"\m": now.strftime("%M"),
+            r"\P": now.strftime("%p"),
+            r"\R": now.strftime("%H"),
+            r"\r": now.strftime("%I"),
+            r"\s": now.strftime("%S"),
+            r"\_": " ",
+        }
+        # Compile a regex pattern that matches any of the keys in replacements
+        pattern = re.compile("|".join(re.escape(key) for key in replacements.keys()))
+
+        # Define the replacement function
+        def replacer(match):
+            return replacements[match.group(0)]
+
+        # Perform the substitution
+        return pattern.sub(replacer, string)
 
     def run_query(self, query, new_line=True):
         """Runs *query*."""
@@ -805,13 +811,7 @@ class LiteCli(object):
         if cur:
             column_types = None
             if hasattr(cur, "description"):
-
-                def get_col_type(col):
-                    # col_type = FIELD_TYPES.get(col[1], text_type)
-                    # return col_type if type(col_type) is type else text_type
-                    return text_type
-
-                column_types = [get_col_type(col) for col in cur.description]
+                column_types = [str(col) for col in cur.description]
 
             if max_width is not None:
                 cur = list(cur)
@@ -824,20 +824,14 @@ class LiteCli(object):
                 **output_kwargs,
             )
 
-            if isinstance(formatted, (text_type)):
+            if isinstance(formatted, str):
                 formatted = formatted.splitlines()
             formatted = iter(formatted)
 
             first_line = next(formatted)
             formatted = itertools.chain([first_line], formatted)
 
-            if (
-                not expanded
-                and max_width
-                and headers
-                and cur
-                and len(first_line) > max_width
-            ):
+            if not expanded and max_width and headers and cur and len(first_line) > max_width:
                 formatted = self.formatter.format_output(
                     cur,
                     headers,
@@ -845,7 +839,7 @@ class LiteCli(object):
                     column_types=column_types,
                     **output_kwargs,
                 )
-                if isinstance(formatted, (text_type)):
+                if isinstance(formatted, str):
                     formatted = iter(formatted.splitlines())
 
             output = itertools.chain(output, formatted)
@@ -865,7 +859,7 @@ class LiteCli(object):
 
 
 @click.command()
-@click.option("-V", "--version", is_flag=True, help="Output litecli's version.")
+@click.version_option(__version__, "-V", "--version")
 @click.option("-D", "--database", "dbname", help="Database to use.")
 @click.option(
     "-R",
@@ -890,19 +884,14 @@ class LiteCli(object):
     is_flag=True,
     help="Automatically switch to vertical output mode if the result is wider than the terminal width.",
 )
-@click.option(
-    "-t", "--table", is_flag=True, help="Display batch output in table format."
-)
+@click.option("-t", "--table", is_flag=True, help="Display batch output in table format.")
 @click.option("--csv", is_flag=True, help="Display batch output in CSV format.")
-@click.option(
-    "--warn/--no-warn", default=None, help="Warn before running a destructive query."
-)
+@click.option("--warn/--no-warn", default=None, help="Warn before running a destructive query.")
 @click.option("-e", "--execute", type=str, help="Execute command and quit.")
 @click.argument("database", default="", nargs=1)
 def cli(
     database,
     dbname,
-    version,
     prompt,
     logfile,
     auto_vertical_output,
@@ -919,11 +908,6 @@ def cli(
       - litecli lite_database
 
     """
-
-    if version:
-        print("Version:", __version__)
-        sys.exit(0)
-
     litecli = LiteCli(
         prompt=prompt,
         logfile=logfile,
@@ -964,10 +948,7 @@ def cli(
         except (FileNotFoundError, OSError):
             litecli.logger.warning("Unable to open TTY as stdin.")
 
-        if (
-            litecli.destructive_warning
-            and confirm_destructive_query(stdin_text) is False
-        ):
+        if litecli.destructive_warning and confirm_destructive_query(stdin_text) is False:
             exit(0)
         try:
             new_line = True
